import re
import typing
from typing import Dict, List, Optional, Tuple, Union

import elasticsearch_dsl as es_dsl

from karp.search.domain import query_dsl  # pyre-ignore
from karp.search.application.queries import QueryRequest
from karp.search.domain.errors import IncompleteQuery, UnsupportedQuery
from karp.search.domain.query import Query
from karp.search.domain.query_dsl import basic_ast as ast
from karp.search.domain.query_dsl import is_a, op


class EsQuery(Query):
    query: typing.Optional[es_dsl.query.Query] = None
    resource_str: typing.Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # self.query = None
        # self.resource_str: Optional[str] = None

    def parse_arguments(self, args, resource_str: str):
        super().parse_arguments(args, resource_str)
        self.resource_str = resource_str
        # if not self.ast.is_empty():
        #     self.query = create_es_query(self.ast.root)

    def _self_name(self) -> str:
        return "EsQuery query={} resource_str={}".format(self.query, self.resource_str)

    @classmethod
    def from_query_request(cls, request: QueryRequest):
        query = cls(fields=[], resources=request.resource_ids, sort=[])
        query.from_ = request.from_
        query.size = request.size
        query.lexicon_stats = request.lexicon_stats
        query.q = request.q or ""
        query.sort = request.sort
        # query.ast = query_dsl.parse(query.q)
        # query._update_ast()
        # if not query.ast.is_empty():
        #     query.query = create_es_query(query.ast.root)
        return query

    class Config:
        arbitrary_types_allowed = True


def get_value(value_node: Union[ast.Node, ast.AnyValue]) -> ast.AnyValue:
    if not isinstance(value_node, ast.Node):
        return value_node
    elif is_a(value_node, op.STRING):
        return value_node.value
    elif is_a(value_node, op.INT):
        return value_node.value
    elif is_a(value_node, op.FLOAT):
        return value_node.value
    else:
        raise IncompleteQuery(
            "Unknown of value (type: {value_node.type}, value: {value_node.value})".format(
                value_node=value_node
            )
        )


def create_es_query(node: ast.Node):
    node.pprint(0)
    if node is None:
        raise TypeError()

    def extract_values(n: ast.Node) -> List[ast.AnyValue]:
        values = []
        if is_a(n, op.ARG_LOGICAL):
            for child in n.children:
                values.append(get_value(child))

        return values

    def extract_values_and_logicals(
        n: ast.Node,
    ) -> Tuple[List[ast.AnyValue], List[ast.Node]]:
        values = []
        logicals = []
        if is_a(n, op.ARG_LOGICAL):
            for child in n.children:
                if is_a(child, op.ARG_LOGICAL):
                    logicals.append(child)
                else:
                    values.append(get_value(child))
        return values, logicals

    def construct_binary_op(field: ast.Node, arg: ast.Node, query_creator):
        q = None
        if is_a(field, op.ARG_LOGICAL):
            if is_a(arg, op.ARG_LOGICAL):
                pass
            else:
                queries = []
                for n in field.children:
                    if is_a(n, op.ARGS):
                        queries.append(query_creator(n, arg))
                    elif is_a(n, op.ARG_LOGICAL):
                        l_queries = []
                        for child in n.children:
                            l_queries.append(query_creator(child, arg))
                        if is_a(n, op.ARG_AND):
                            queries.append(es_dsl.Q("bool", must=l_queries))
                        elif is_a(n, op.ARG_OR):
                            queries.append(es_dsl.Q("bool", should=l_queries))
                        else:
                            queries.append(es_dsl.Q("bool", must_not=l_queries))
                if is_a(arg, op.ARG_AND):
                    q = es_dsl.Q("bool", must=queries)
                elif is_a(arg, op.ARG_OR):
                    q = es_dsl.Q("bool", should=queries)
                else:
                    q = es_dsl.Q("bool", must_not=queries)
                if not arg_values:

                    def prepare_equals_arg(node, v):
                        return v

                    arg = prepare_equals_arg(node, get_value(arg2))
                    queries = [
                        construct_equals_query(field, arg) for field in field_values
                    ]
                    for logical in field_logicals:
                        l_queries = []
                        for field in logical.children:
                            l_queries.append(construct_equals_query(field.value, arg))
                        if is_a(logical, op.ARG_OR):
                            queries.append(es_dsl.Q("bool", should=l_queries))
                        elif is_a(logical, op.ARG_AND):
                            queries.append(es_dsl.Q("bool", must=l_queries))
                        else:
                            queries.append(es_dsl.Q("bool", must_not=l_queries))
                    print(
                        "|create_es_query::EQUALS| queries = {queries}".format(
                            queries=queries
                        )
                    )
                    if is_a(arg1, op.ARG_AND):
                        q = es_dsl.Q("bool", must=queries)
                    elif is_a(arg1, op.ARG_OR):
                        q = es_dsl.Q("bool", should=queries)
                    else:
                        q = es_dsl.Q(
                            "bool",
                            must_not=queries,
                            must=es_dsl.Q("multi_match", query=get_value(arg2)),
                        )
        else:
            if is_a(arg, op.ARG_LOGICAL):
                queries = []
                for n in arg.children:
                    if is_a(n, op.ARGS):
                        queries.append(query_creator(field, n))
                    elif is_a(n, op.ARG_LOGICAL):
                        l_queries = []
                        for child in n.children:
                            l_queries.append(query_creator(field, child))
                        if is_a(n, op.ARG_AND):
                            queries.append(es_dsl.Q("bool", must=l_queries))
                        elif is_a(n, op.ARG_OR):
                            queries.append(es_dsl.Q("bool", should=l_queries))
                        else:
                            queries.append(es_dsl.Q("bool", must_not=l_queries))
                if is_a(arg, op.ARG_AND):
                    q = es_dsl.Q("bool", must=queries)
                elif is_a(arg, op.ARG_OR):
                    q = es_dsl.Q("bool", should=queries)
                else:
                    q = es_dsl.Q("bool", must_not=queries)
            else:
                q = query_creator(field, arg)
        return q

    def construct_equals_query(field: ast.Node, query: ast.Node):
        kwargs = {get_value(field): {"query": get_value(query), "operator": "and"}}
        return es_dsl.Q("match", **kwargs)

    q = None
    if is_a(node, op.LOGICAL):
        # TODO check minimum should match rules in different contexts
        queries = [create_es_query(n) for n in node.children]
        if len(queries) == 2:
            q1 = queries[0]
            q2 = queries[1]
            print("q1 = {}".format(q1.to_dict()))
            print("q2 = {}".format(repr(q2)))
            q1_dict = q1.to_dict()
            q2_dict = q2.to_dict()
            if "range" in q1_dict and "range" in q2_dict:
                for q1_field, q1_value in q1_dict["range"].items():
                    for q2_field, q2_value in q2_dict["range"].items():
                        if q1_field == q2_field:
                            print("q1_field == q2_field")
                            range_args = q1_value
                            range_args.update(q2_value)
                            print("field = {}".format(q1_field))
                            print("range_args = {}".format(range_args))
                            q = es_dsl.Q("range", **{q1_field: range_args})
                            return q
        if is_a(node, op.AND):
            q = es_dsl.Q("bool", must=queries)
        elif is_a(node, op.OR):
            q = es_dsl.Q("bool", should=queries)
        else:
            q = es_dsl.Q("bool", must_not=queries)
    elif is_a(node, op.UNARY_OPS):
        arg = node.children[0]
        arg_values, arg_logicals = extract_values_and_logicals(arg)
        print("arg_values = {}".format(arg_values))
        print("arg_logicals = {}".format(arg_logicals))

        def construct_unary_op(arg: ast.Node, query_creator):
            q = None
            if is_a(arg, op.ARG_LOGICAL):
                queries = []
                for n in arg.children:
                    if is_a(n, op.ARGS):
                        queries.append(query_creator(n))
                    elif is_a(n, op.ARG_LOGICAL):
                        l_queries = []
                        for child in n.children:
                            l_queries.append(query_creator(child))
                        if is_a(n, op.ARG_AND):
                            queries.append(es_dsl.Q("bool", must=l_queries))
                        elif is_a(n, op.ARG_OR):
                            queries.append(es_dsl.Q("bool", should=l_queries))
                        else:
                            queries.append(es_dsl.Q("bool", must_not=l_queries))
                if is_a(arg, op.ARG_AND):
                    q = es_dsl.Q("bool", must=queries)
                elif is_a(arg, op.ARG_OR):
                    q = es_dsl.Q("bool", should=queries)
                else:
                    q = es_dsl.Q("bool", must_not=queries)
            else:
                q = query_creator(arg)
            return q

        def construct_exists_query(node: ast.Node):
            return es_dsl.Q("exists", field=get_value(node))

        if is_a(node, op.FREETEXT):

            def construct_freetext_query(node):
                if is_a(node, op.STRING):
                    return es_dsl.Q("multi_match", query=get_value(node), fuzziness=1)
                else:
                    return es_dsl.Q("multi_match", query=get_value(node))

            if not arg_values:
                q = construct_freetext_query(arg)
            else:
                queries = [construct_freetext_query(n) for n in arg.children]
                if is_a(arg, op.ARG_OR):
                    q = es_dsl.Q("bool", should=queries)
                elif is_a(arg, op.ARG_AND):
                    q = es_dsl.Q("bool", must=queries)
                else:
                    q = es_dsl.Q("bool", must_not=queries)
        elif is_a(node, op.FREERGXP):
            kwargs = {"default_field": "*"}
            if not arg_values:
                kwargs["query"] = "/{}/".format(arg.value)
            else:
                if is_a(arg, op.ARG_OR):
                    operator = " OR "
                else:  # if is_a(arg, op.ARG_AND):
                    operator = " AND "
                kwargs["query"] = operator.join("(/{}/)".format(v) for v in arg_values)
            print("kwargs = {}".format(kwargs))
            q = es_dsl.Q("query_string", **kwargs)
            if is_a(arg, op.ARG_NOT):
                q = es_dsl.Q("bool", must_not=q)
        elif is_a(node, op.EXISTS):
            q = construct_unary_op(node.children[0], construct_exists_query)

        elif is_a(node, op.MISSING):
            if not arg_values:
                q = es_dsl.Q("bool", must_not=es_dsl.Q("exists", field=arg.value))
            else:
                queries = [es_dsl.Q("exists", field=value) for value in arg_values]
                for logical in arg_logicals:
                    l_queries = []
                    for field in logical.children:
                        l_queries.append(construct_exists_query(field))
                    if is_a(logical, op.ARG_AND):
                        queries.append(es_dsl.Q("bool", must=l_queries))
                    elif is_a(logical, op.ARG_OR):
                        queries.append(es_dsl.Q("bool", should=l_queries))
                    else:  # if is_a(logical, op.ARG_NOT):
                        queries.append(es_dsl.Q("bool", must_not=l_queries))

                if is_a(arg, op.ARG_AND):
                    q = es_dsl.Q("bool", must_not=queries)
                elif is_a(arg, op.ARG_OR):
                    q = None
                    for query in queries:
                        q_tmp = es_dsl.Q("bool", must_not=query)
                        if not q:
                            q = q_tmp
                        else:
                            q = q | q_tmp
                else:  # is_a(arg, op.ARG_NOT):
                    q = es_dsl.Q("bool", must=queries)
        else:
            raise UnsupportedQuery("not implemented")
    elif is_a(node, op.BINARY_OPS):
        arg1 = node.children[0]
        arg2 = node.children[1]
        field_values, field_logicals = extract_values_and_logicals(arg1)
        arg_values = extract_values(arg2)

        print("field_values = {field_values}".format(field_values=field_values))
        print("field_logicals = {field_logicals}".format(field_logicals=field_logicals))

        print("arg_values = {}".format(arg_values))
        # TODO this check breaks and and or since they always (?) have ast.ArgNode as parameters
        # if not isinstance(arg1, ast.ArgNode) or not isinstance(arg2, ast.ArgNode):
        # TODO these need to be moved outside of current query, for example:
        # "equals|name||or|Partille|Kumla" could be expressed in two ways
        # es_dsl.Q('terms', name=['Partille', 'Kumla'])
        # or
        # es_dsl.Q('bool', should=[es_dsl.Q('term', name='Partille'), es_dsl.Q('term', name='Kumla')])
        # but "regexp|name||or|Part*|Kum*"
        # can only be expressed as in the longer form above
        # raise UnsupportedQuery()

        if is_a(node, op.EQUALS):

            # q = construct_binary_op(node.children[0], node.children[1], construct_equals_query)
            # if len(field_values) == 1:
            if not field_values and not field_logicals:
                if not arg_values:
                    q = construct_equals_query(arg1, arg2)
                else:
                    queries = [
                        construct_equals_query(get_value(arg1), query)
                        for query in arg_values
                    ]
                    if is_a(arg2, op.ARG_AND):
                        q = es_dsl.Q("bool", must=queries)
                    elif is_a(arg2, op.ARG_OR):
                        q = es_dsl.Q("bool", should=queries)
                    else:
                        q = es_dsl.Q(
                            "bool", must_not=queries
                        )  # , must=es_dsl.Q('multi_match', query=get_value(arg1)))

            else:  # if field_values:
                # if len(arg_values) == 1:
                if not arg_values:

                    def prepare_equals_arg(node, v):
                        return v

                    arg = prepare_equals_arg(node, get_value(arg2))
                    queries = [
                        construct_equals_query(field, arg) for field in field_values
                    ]
                    for logical in field_logicals:
                        l_queries = []
                        for field in logical.children:
                            l_queries.append(construct_equals_query(field.value, arg))
                        if is_a(logical, op.ARG_OR):
                            queries.append(es_dsl.Q("bool", should=l_queries))
                        elif is_a(logical, op.ARG_AND):
                            queries.append(es_dsl.Q("bool", must=l_queries))
                        else:
                            queries.append(es_dsl.Q("bool", must_not=l_queries))
                    print(
                        "|create_es_query::EQUALS| queries = {queries}".format(
                            queries=queries
                        )
                    )
                    if is_a(arg1, op.ARG_AND):
                        q = es_dsl.Q("bool", must=queries)
                    elif is_a(arg1, op.ARG_OR):
                        q = es_dsl.Q("bool", should=queries)
                    else:
                        q = es_dsl.Q(
                            "bool",
                            must_not=queries,
                            must=es_dsl.Q("multi_match", query=get_value(arg2)),
                        )
                else:  # if arg_values:
                    raise UnsupportedQuery("Don't know how to handle ")

        elif is_a(node, op.REGEX_OPS):

            def extract_field_plus_rawfield(field: str):
                if field.endswith(".raw"):
                    yield field
                else:
                    yield field
                    yield field + ".raw"

            def prepare_regex(node, s: str):
                if is_a(node, op.CONTAINS):
                    return ".*" + re.escape(s) + ".*"
                elif is_a(node, op.STARTSWITH):
                    return re.escape(s) + ".*"
                elif is_a(node, op.ENDSWITH):
                    return ".*" + re.escape(s)
                else:
                    return s

            # Construct query
            def construct_regexp_query(field_in: str, regex: str):
                q = None
                for field in extract_field_plus_rawfield(field_in):
                    q_tmp = es_dsl.Q("regexp", **{field: regex})
                    if not q:
                        q = q_tmp
                    else:
                        q = q | q_tmp
                return q

            q = None
            if not field_values and not field_logicals:
                if not arg_values:
                    q = construct_regexp_query(
                        get_value(arg1), prepare_regex(node, get_value(arg2))
                    )
                else:
                    queries = []
                    for regex in arg_values:
                        q_tmp = construct_regexp_query(
                            get_value(arg1), prepare_regex(node, regex)
                        )
                        print("q_tmp = {q_tmp}".format(q_tmp=q_tmp))

                        queries.append(q_tmp)

                    if is_a(arg2, op.ARG_OR):
                        q = es_dsl.Q("bool", should=queries)
                    elif is_a(arg2, op.ARG_AND):
                        q = es_dsl.Q("bool", must=queries)
                    else:
                        q = es_dsl.Q(
                            "bool", must_not=queries
                        )  # , must=es_dsl.Q('query_string', **kwargs))
            else:  # if field_values:
                if not arg_values:
                    regex = prepare_regex(node, get_value(arg2))
                    queries = []
                    for field in field_values:
                        q_tmp = construct_regexp_query(field, regex)
                        print("q_tmp = {q_tmp}".format(q_tmp=q_tmp))

                        queries.append(q_tmp)
                    for logical in field_logicals:
                        l_queries = []
                        for field in logical.children:
                            l_queries.append(construct_regexp_query(field.value, regex))
                        if is_a(logical, op.ARG_OR):
                            queries.append(es_dsl.Q("bool", should=l_queries))
                        elif is_a(logical, op.ARG_AND):
                            queries.append(es_dsl.Q("bool", must=l_queries))
                        else:
                            queries.append(es_dsl.Q("bool", must_not=l_queries))

                    print(
                        "|create_es_query::REGEX| queries = {queries}".format(
                            queries=queries
                        )
                    )
                    if is_a(arg1, op.ARG_OR):
                        q = es_dsl.Q("bool", should=queries)
                    elif is_a(arg1, op.ARG_AND):
                        q = es_dsl.Q("bool", must=queries)
                    else:
                        kwargs = {"default_field": "*"}
                        kwargs["query"] = "/{}/".format(regex)

                        print("regex NOT kwargs = {}".format(kwargs))

                        q = es_dsl.Q(
                            "bool",
                            must_not=queries,
                            must=es_dsl.Q("query_string", **kwargs),
                        )
                else:
                    raise UnsupportedQuery("Complex regex not implemented")

        elif is_a(node, op.RANGE_OPS):
            if arg_values:
                raise UnsupportedQuery(
                    "Not allowed to use logical operators in 2nd argument for RANGE operators."
                )

            def prepare_range_args(node, arg22):
                range_args = {}
                if is_a(node, op.LT):
                    range_args["lt"] = arg22
                elif is_a(node, op.LTE):
                    range_args["lte"] = arg22
                elif is_a(node, op.GT):
                    range_args["gt"] = arg22
                elif is_a(node, op.GTE):
                    range_args["gte"] = arg22
                return range_args

            range_args = prepare_range_args(node, get_value(arg2))

            def construct_range_query(field: str, range_args: Dict):
                return es_dsl.Q("range", **{field: range_args})

            q = None
            if not field_values:
                q = construct_range_query(get_value(arg1), range_args)
            else:
                queries = [construct_range_query(f, range_args) for f in field_values]
                for logical in field_logicals:
                    l_queries = []
                    for field in logical.children:
                        l_queries.append(construct_range_query(field.value, range_args))
                    if is_a(logical, op.ARG_OR):
                        queries.append(es_dsl.Q("bool", should=l_queries))
                    elif is_a(logical, op.ARG_AND):
                        queries.append(es_dsl.Q("bool", must=l_queries))
                    else:
                        queries.append(es_dsl.Q("bool", must_not=l_queries))
                print(
                    "|create_es_query::RANGE| queries = {queries}".format(
                        queries=queries
                    )
                )
                if is_a(arg1, op.ARG_AND):
                    q = es_dsl.Q("bool", must=queries)
                elif is_a(arg1, op.ARG_OR):
                    q = es_dsl.Q("bool", should=queries)
                else:
                    q = es_dsl.Q("bool", must_not=queries)
    # elif isinstance(node, ast.TernaryOp):
    #     op = node.value
    #     arg1 = node.children[0]
    #     arg2 = node.children[1]
    #     arg3 = node.children[2]
    #     if op == query_dsl.Operators.RANGE:
    #         raise UnsupportedQuery('don\'t now what to do yet')
    #     else:
    #         raise UnsupportedQuery('what operators?')
    else:
        raise UnsupportedQuery("Unknown query op '{node}'".format(node=node))

    return q


# def parse_sortable_fields(properties: Dict[str, Any]) -> Dict[str, List[str]]:
#     for prop_name, prop_value in properties.items():
#         if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
#             return [prop_name]



class InMemoryUnitOfWork:
    def __init__(self, **kwargs):
        self.was_committed = False
        self.was_rolled_back = False
        self.was_closed = False

    def start(self):
        self.was_committed = False
        self.was_rolled_back = False
        self.was_closed = False
        return self

    def __enter__(self):
        self.was_committed = False
        self.was_rolled_back = False
        return self

    def _commit(self):
        self.was_committed = True

    def rollback(self):
        self.was_rolled_back = True

    def _close(self):
        self.was_closed = True



"""Pytest entry point."""

# pylint: disable=wrong-import-position,missing-function-docstring
from karp.main import AppContext
from karp.tests.integration.auth.adapters import create_bearer_token
from karp import auth, config
import os
import json
import typing
from typing import Any, Dict, Generator, Optional, Tuple

from fastapi import FastAPI
from typer import Typer
from typer.testing import CliRunner

import alembic
import alembic.config
from karp.foundation.value_objects import make_unique_id
from karp.foundation.commands import CommandBus
import elasticsearch_test  # pyre-ignore

import pytest  # pyre-ignore
from sqlalchemy import create_engine, pool
from sqlalchemy.orm import session, sessionmaker
from starlette.testclient import TestClient

from alembic.config import main as alembic_main

# environ["TESTING"] = "True"
# environ["ELASTICSEARCH_HOST"] = "localhost:9202"
# environ["CONSOLE_LOG_LEVEL"] = "DEBUG"

# print("importing karp stuf ...")
from karp.tests import common_data, utils  # nopep8
from karp.auth_infrastructure import TestAuthInfrastructure  # nopep8
import karp.lex_infrastructure.sql.sql_models  # nopep8
from karp.db_infrastructure.db import metadata  # nopep8
from karp.lex.domain import commands, errors, entities  # nopep8
from karp import errors as karp_errors  # nopep8


@pytest.fixture(name="in_memory_sqlite_db")
def fixture_in_memory_sqlite_db():
    engine = create_engine("sqlite:///:memory:")
    metadata.create_all(engine)
    yield engine
    session.close_all_sessions()
    metadata.drop_all(bind=engine)


@pytest.fixture
def sqlite_session_factory(in_memory_sqlite_db):
    yield sessionmaker(bind=in_memory_sqlite_db)


@pytest.fixture(scope="session")
def setup_environment() -> None:
    os.environ["TESTING"] = "1"
    os.environ["AUTH_JWT_PUBKEY_PATH"] = "karp/tests/data/pubkey.pem"
    os.environ["ELASTICSEARCH_HOST"] = "localhost:9202"


@pytest.fixture(scope="session")
def apply_migrations(setup_environment: None):
    from karp.main.migrations import use_cases

    print("running alembic upgrade ...")
    uc = use_cases.RunningMigrationsUp()
    uc.execute(use_cases.RunMigrationsUp())
    yield


@pytest.fixture(scope="session", name="runner")
def fixture_runner() -> CliRunner:
    return CliRunner()


@pytest.fixture(scope="session", name="cliapp")
def fixture_cliapp() -> Typer:
    from karp.cliapp.main import create_app

    return create_app()


@pytest.fixture(name="app", scope="session")
def fixture_app(
    apply_migrations: None, init_search_service: None
) -> Generator[FastAPI, None, None]:
    from karp.webapp.main import create_app

    app = create_app()
    # app.state.app_context.container.binder.install(TestAuthInfrastructure())
    yield app
    print("dropping app")
    app = None


@pytest.fixture(name="app_context", scope="session")
def fixture_app_context(app: FastAPI) -> AppContext:
    return app.state.app_context


@pytest.fixture(name="fa_client", scope="session")
def fixture_client(app: FastAPI) -> Generator[TestClient, None, None]:
    with TestClient(app) as client:
        yield client


#     async with LifespanManager(app):
#         async with AsyncClient(
#             app=app,
#             base_url="http://testserver",
#             headers={"Content-Type": "application/json"}
#         ) as client:
#             yield client


def create_and_publish_resource(
    client: TestClient,
    *,
    path_to_config: str,
    access_token: auth.AccessToken,
) -> Tuple[bool, Optional[dict[str, Any]]]:
    from karp.webapp.schemas import ResourceCreate

    with open(path_to_config) as fp:
        resource_config = json.load(fp)

    resource_id = resource_config.pop("resource_id")
    resource = ResourceCreate(
        resource_id=resource_id,
        name=resource_config.pop("resource_name"),
        config=resource_config,
        message=f"{resource_id} added",
    )
    response = client.post(
        "/resources/",
        json=resource.dict(),
        headers=access_token.as_header(),
    )
    if response.status_code != 201:
        return False, {
            "error": f"Failed create resource: unexpected status_code: {response.status_code} != 201 ",
            "response.json": response.json(),
        }

    response = client.post(
        f"/resources/{resource_id}/publish",
        json={
            "message": f"{resource_id} published",
            "version": 1,
        },
        headers=access_token.as_header(),
    )
    if response.status_code != 200:
        return False, {
            "error": f"Failed publish resource: unexpected status_code: {response.status_code} != 200 ",
            "response.json": response.json(),
        }

    return True, None


@pytest.fixture(scope="session", name="fa_data_client")
def fixture_fa_data_client(
    fa_client,
    admin_token: auth.AccessToken,
):
    ok, msg = create_and_publish_resource(
        fa_client,
        path_to_config="karp/tests/data/config/places.json",
        access_token=admin_token,
    )
    assert ok, msg
    ok, msg = create_and_publish_resource(
        fa_client,
        path_to_config="karp/tests/data/config/municipalities.json",
        access_token=admin_token,
    )
    assert ok, msg
    utils.add_entries(
        fa_client,
        {"places": common_data.PLACES, "municipalities": common_data.MUNICIPALITIES},
        access_token=admin_token,
    )

    return fa_client


@pytest.fixture(scope="session")
def auth_levels() -> typing.Dict[str, int]:
    curr_level = 10
    levels = {}
    for level in auth.PermissionLevel:
        levels[level.value] = curr_level
        curr_level += 10

    return levels


@pytest.fixture(scope="session")
def user1_token(auth_levels: typing.Dict[str, int]) -> auth.AccessToken:
    return create_bearer_token(
        user="user1",
        levels=auth_levels,
        scope={
            "lexica": {
                "places": auth_levels[auth.PermissionLevel.write],
            }
        },
    )


@pytest.fixture(scope="session")
def user2_token(auth_levels: typing.Dict[str, int]) -> auth.AccessToken:
    return create_bearer_token(
        user="user2",
        levels=auth_levels,
        scope={
            "lexica": {
                "places": auth_levels[auth.PermissionLevel.write],
            }
        },
    )


@pytest.fixture(scope="session")
def user4_token(auth_levels: typing.Dict[str, int]) -> auth.AccessToken:
    return create_bearer_token(
        user="user4",
        levels=auth_levels,
        scope={
            "lexica": {
                "places": auth_levels[auth.PermissionLevel.admin],
            }
        },
    )


@pytest.fixture(scope="session")
def admin_token(auth_levels: typing.Dict[str, int]) -> auth.AccessToken:
    return create_bearer_token(
        user="alice@example.com",
        levels=auth_levels,
        scope={
            "lexica": {
                "places": auth_levels[auth.PermissionLevel.admin],
                "test_resource": auth_levels[auth.PermissionLevel.admin],
                "municipalities": auth_levels[auth.PermissionLevel.admin],
            }
        },
    )


@pytest.fixture(scope="session")
def read_token(auth_levels: typing.Dict[str, int]) -> auth.AccessToken:
    return create_bearer_token(
        user="bob@example.com",
        levels=auth_levels,
        scope={
            "lexica": {
                "places": auth_levels[auth.PermissionLevel.read],
                "test_resource": auth_levels[auth.PermissionLevel.read],
                "municipalities": auth_levels[auth.PermissionLevel.read],
            }
        },
    )


@pytest.fixture(scope="session")
def write_token(auth_levels: typing.Dict[str, int]) -> auth.AccessToken:
    return create_bearer_token(
        user="charlie@example.com",
        levels=auth_levels,
        scope={
            "lexica": {
                "places": auth_levels[auth.PermissionLevel.write],
                "test_resource": auth_levels[auth.PermissionLevel.write],
                "municipalities": auth_levels[auth.PermissionLevel.write],
            }
        },
    )


@pytest.fixture(name="init_search_service", scope="session")
def fixture_init_search_service():
    print("fixture: use_main_index")
    if not config.TEST_ELASTICSEARCH_ENABLED:
        print("don't use elasticsearch")
        # pytest.skip()
        yield
    else:
        if not config.TEST_ES_HOME:
            raise RuntimeError("must set ES_HOME to run tests that use elasticsearch")
        es_port = int(os.environ.get("TEST_ELASTICSEARCH_PORT", "9202"))
        with elasticsearch_test.ElasticsearchTest(
            port=es_port, es_path=config.TEST_ES_HOME
        ):
            yield


@pytest.fixture
def json_schema_config():
    return common_data.CONFIG_PLACES


# #
# #
# # @pytest.fixture(scope="session")
# # def client_with_entries_scope_session(es, client_with_data_f_scope_session):
# #     client_with_data = init(
# #         client_with_data_f_scope_session,
# #         es,
# #         {"places": PLACES, "municipalities": MUNICIPALITIES},
# #     )
# #     time.sleep(5)
# #     return client_with_data



CONFIG_PLACES = {
    "resource_id": "places",
    "resource_name": "Platser i Sverige",
    "fields": {
        "name": {"type": "string", "required": True},
        "municipality": {"collection": True, "type": "number", "required": True},
        "population": {"type": "number"},
        "area": {"type": "number"},
        "density": {"type": "number"},
        "code": {"type": "number", "required": True},
    },
    "sort": "name",
    "id": "code",
    "referenceable": ["name", "municipality", "code"],
}


PLACES = [
    {
        "code": 1,
        "name": "Grund test",
        "population": 3122,
        "area": 6312,
        "density": 6312,
        "municipality": [1],
        "larger_place": 7  # Alhamn
        # "smaller_places": 9 "Bjurvik2"
    },
    {
        "code": 2,
        "name": "Grunds",
        "population": 6312,
        "area": 20000,
        "density": 6,
        "municipality": [1],
    },
    {
        "code": 3,
        "name": "Botten test",
        "population": 4133,
        "area": 50000,
        "density": 7,
        "municipality": [2, 3],
        "larger_place": 8  # "Bjurvik"
        # "smaller_places": 4 "Hambo"
    },
    {
        "code": 4,
        "name": "Hambo",
        "population": 4132,
        "area": 50000,
        "municipality": [2, 3],
        "larger_place": 3  # Botten test
        # "smaller_places": 7 "Alhamn"
    },
    {"code": 5, "name": "Rutvik", "area": 50000, "municipality": [2, 3]},
    {
        "code": 6,
        "name": "Alvik",
        "area": 6312,
        "population": 6312,
        "density": 12,
        "municipality": [3]
        # "smaller_places": 7  "Bjurvik"
    },
    {
        "code": 7,
        "name": "Alhamn",
        "area": 6312,
        "population": 3812,
        "density": 12,
        "municipality": [2, 3],
        "larger_place": 4  # Hambo
        # "smaller_places": 1  "Grund test"
    },
    {
        "code": 8,
        "name": "Bjurvik",
        "area": 6312,
        "population": 6212,
        "density": 12,
        "municipality": [2, 3],
        "larger_place": 6  # Alvik
        # "smaller_places": 1  "Botten test"
    },
    {
        "code": 9,
        "name": "Bjurvik2",
        "area": 7312,
        "population": 641,
        "density": 12,
        "municipality": [2],
        "larger_place": 1,  # Grund test
    },
]


MUNICIPALITIES = [
    {
        "code": 1,
        "name": "Luleå kommun",
        "state": "Norrbottens län",
        "region": "Norrbotten",
        "capital": "Luleå",
        "area": {"land": 2094.08, "water": 148.39},
        "population": {"value": {"total": 77860}, "density": {"total": 37.18}},
    },
    {
        "code": 2,
        "name": "Norsjö kommun",
        "state": "Västerbottens län",
        "region": "Västerbotten",
        "capital": "Norsjö",
        "area": {"land": 1739.15, "water": 184.43},
        "population": {"value": {"total": 4101}, "density": {"total": 2.36}},
    },
    {
        "code": 3,
        "name": "Piteå kommun",
        "state": "Norrbottens län",
        "region": "Norrbotten",
        "capital": "Piteå",
        "area": {"land": 3086.04, "water": 149.04},
        "population": {"value": {"total": 42108}, "density": {"total": 13.64}},
    },
]
